"""Test root services on supervisor client."""

from json import dumps

from aiohttp import ClientSession
from aioresponses import aioresponses
import pytest
from yarl import URL

from aiohasupervisor import (
    SupervisorAuthenticationError,
    SupervisorBadRequestError,
    SupervisorClient,
    SupervisorError,
    SupervisorForbiddenError,
    SupervisorNotFoundError,
    SupervisorServiceUnavailableError,
)
from aiohasupervisor.models import HostFeature, SupervisorState, UpdateType

from . import load_fixture
from .const import SUPERVISOR_URL


async def test_using_own_session(responses: aioresponses) -> None:
    """Test passing in an existing session."""
    responses.get(
        f"{SUPERVISOR_URL}/info", status=200, body=load_fixture("root_info.json")
    )
    async with ClientSession() as session:
        client = SupervisorClient(SUPERVISOR_URL, "abc123", session=session)
        info = await client.info()
        assert info
        assert client._client.session == session
        assert not session.closed
        await client.close()
        assert not session.closed


async def test_using_new_session(responses: aioresponses) -> None:
    """Test letting client create new session."""
    responses.get(
        f"{SUPERVISOR_URL}/info", status=200, body=load_fixture("root_info.json")
    )
    async with SupervisorClient(SUPERVISOR_URL, "abc123") as client:
        info = await client.info()
        assert info
        assert client._client.session
        assert not client._client.session.closed

    assert client._client.session.closed


async def test_info(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test info API."""
    responses.get(
        f"{SUPERVISOR_URL}/info", status=200, body=load_fixture("root_info.json")
    )
    info = await supervisor_client.info()
    assert info.supervisor == "2024.07.1.dev3001"
    assert info.homeassistant == "2024.6.0.dev202405280218"
    assert info.hassos == "12.4.dev20240527"
    assert info.hostname == "homeassistant"
    assert info.arch == "aarch64"
    assert info.supported is True
    assert info.state == SupervisorState.RUNNING
    assert HostFeature.REBOOT in info.features
    assert "not_real" in info.features


async def test_available_updates(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test available updates API."""
    responses.get(
        f"{SUPERVISOR_URL}/available_updates",
        status=200,
        body=load_fixture("root_available_updates.json"),
    )
    updates = await supervisor_client.available_updates()
    assert updates[0].update_type == UpdateType.CORE
    assert updates[0].panel_path == "/update-available/core"
    assert updates[0].version_latest == "2024.9.0.dev202408010224"
    assert updates[0].name is None
    assert updates[0].icon is None
    assert updates[1].update_type == UpdateType.OS


async def test_refresh_updates(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test refresh updates API."""
    responses.post(f"{SUPERVISOR_URL}/refresh_updates", status=200)
    assert await supervisor_client.refresh_updates() is None
    assert responses.requests.keys() == {
        ("POST", URL(f"{SUPERVISOR_URL}/refresh_updates"))
    }


@pytest.mark.parametrize(
    ("status", "message", "expected_exc"),
    [
        (400, "Bad input", SupervisorBadRequestError),
        (401, None, SupervisorAuthenticationError),
        (403, "Not allowed", SupervisorForbiddenError),
        (404, "Not found", SupervisorNotFoundError),
        (503, "DB migration in progress", SupervisorServiceUnavailableError),
        (500, "Unknown error", SupervisorError),
    ],
)
async def test_error_handling(
    responses: aioresponses,
    supervisor_client: SupervisorClient,
    status: int,
    message: str | None,
    expected_exc: type[SupervisorError],
) -> None:
    """Test error handling scenarios."""
    if message is not None:
        kwargs = {
            "body": dumps(
                {
                    "result": "error",
                    "message": message,
                }
            )
        }
    else:
        kwargs = {"content_type": "text/plain"}

    responses.post(f"{SUPERVISOR_URL}/refresh_updates", status=status, **kwargs)

    with pytest.raises(expected_exc, match=message):
        await supervisor_client.refresh_updates()



"""Backups client for supervisor."""

from collections.abc import AsyncIterator

from aiohttp import MultipartWriter
from multidict import MultiDict

from .client import _SupervisorComponentClient
from .const import ResponseType
from .models.backups import (
    Backup,
    BackupComplete,
    BackupJob,
    BackupList,
    BackupsInfo,
    BackupsOptions,
    DownloadBackupOptions,
    FreezeOptions,
    FullBackupOptions,
    FullRestoreOptions,
    NewBackup,
    PartialBackupOptions,
    PartialRestoreOptions,
    RemoveBackupOptions,
    UploadBackupOptions,
    UploadedBackup,
)


class BackupsClient(_SupervisorComponentClient):
    """Handles backups access in Supervisor."""

    async def list(self) -> list[Backup]:
        """List backups."""
        result = await self._client.get("backups")
        return BackupList.from_dict(result.data).backups

    async def info(self) -> BackupsInfo:
        """Get backups info."""
        result = await self._client.get("backups/info")
        return BackupsInfo.from_dict(result.data)

    async def set_options(self, options: BackupsOptions) -> None:
        """Set options for backups."""
        await self._client.post("backups/options", json=options.to_dict())

    async def reload(self) -> None:
        """Reload backups cache."""
        await self._client.post("backups/reload")

    async def freeze(self, options: FreezeOptions | None = None) -> None:
        """Start a freeze for external snapshot process."""
        await self._client.post(
            "backups/freeze", json=options.to_dict() if options else None
        )

    async def thaw(self) -> None:
        """Thaw an active freeze when external snapshot process ends."""
        await self._client.post("backups/thaw")

    async def full_backup(self, options: FullBackupOptions | None = None) -> NewBackup:
        """Create a new full backup."""
        result = await self._client.post(
            "backups/new/full",
            json=options.to_dict() if options else None,
            response_type=ResponseType.JSON,
            timeout=None,
        )
        return NewBackup.from_dict(result.data)

    async def partial_backup(self, options: PartialBackupOptions) -> NewBackup:
        """Create a new partial backup."""
        result = await self._client.post(
            "backups/new/partial",
            json=options.to_dict(),
            response_type=ResponseType.JSON,
            timeout=None,
        )
        return NewBackup.from_dict(result.data)

    async def backup_info(self, backup: str) -> BackupComplete:
        """Get backup details."""
        result = await self._client.get(f"backups/{backup}/info")
        return BackupComplete.from_dict(result.data)

    async def remove_backup(
        self, backup: str, options: RemoveBackupOptions | None = None
    ) -> None:
        """Remove a backup."""
        await self._client.delete(
            f"backups/{backup}", json=options.to_dict() if options else None
        )

    async def full_restore(
        self, backup: str, options: FullRestoreOptions | None = None
    ) -> BackupJob:
        """Start full restore from backup."""
        result = await self._client.post(
            f"backups/{backup}/restore/full",
            json=options.to_dict() if options else None,
            response_type=ResponseType.JSON,
            timeout=None,
        )
        return BackupJob.from_dict(result.data)

    async def partial_restore(
        self, backup: str, options: PartialRestoreOptions
    ) -> BackupJob:
        """Start partial restore from backup."""
        result = await self._client.post(
            f"backups/{backup}/restore/partial",
            json=options.to_dict(),
            response_type=ResponseType.JSON,
            timeout=None,
        )
        return BackupJob.from_dict(result.data)

    async def upload_backup(
        self, stream: AsyncIterator[bytes], options: UploadBackupOptions | None = None
    ) -> str:
        """Upload backup by stream and return slug."""
        params = MultiDict()
        if options and options.location:
            for location in options.location:
                params.add("location", location or "")

        with MultipartWriter("form-data") as mp:
            mp.append(stream)
            result = await self._client.post(
                "backups/new/upload",
                params=params,
                data=mp,
                response_type=ResponseType.JSON,
            )

        return UploadedBackup.from_dict(result.data).slug

    async def download_backup(
        self, backup: str, options: DownloadBackupOptions | None = None
    ) -> AsyncIterator[bytes]:
        """Download backup and return stream."""
        params = MultiDict()
        if options and options.location:
            params.add("location", options.location or "")

        result = await self._client.get(
            f"backups/{backup}/download",
            params=params,
            response_type=ResponseType.STREAM,
        )
        return result.data



"""Test for supervisor management client."""

from ipaddress import IPv4Address

from aioresponses import aioresponses
import pytest
from yarl import URL

from aiohasupervisor import SupervisorClient
from aiohasupervisor.models import SupervisorOptions, SupervisorUpdateOptions

from . import load_fixture
from .const import SUPERVISOR_URL


async def test_supervisor_ping(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test supervisor ping API."""
    responses.get(f"{SUPERVISOR_URL}/supervisor/ping", status=200)
    assert await supervisor_client.supervisor.ping() is None
    assert responses.requests.keys() == {
        ("GET", URL(f"{SUPERVISOR_URL}/supervisor/ping"))
    }


async def test_supervisor_info(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test supervisor info API."""
    responses.get(
        f"{SUPERVISOR_URL}/supervisor/info",
        status=200,
        body=load_fixture("supervisor_info.json"),
    )
    info = await supervisor_client.supervisor.info()

    assert info.version == "2024.09.1"
    assert info.channel == "stable"
    assert info.arch == "aarch64"
    assert info.supported is True
    assert info.healthy is True
    assert info.logging == "info"
    assert info.ip_address == IPv4Address("172.30.32.2")


async def test_supervisor_stats(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test supervisor stats API."""
    responses.get(
        f"{SUPERVISOR_URL}/supervisor/stats",
        status=200,
        body=load_fixture("supervisor_stats.json"),
    )
    stats = await supervisor_client.supervisor.stats()

    assert stats.cpu_percent == 0.04
    assert stats.memory_usage == 243982336
    assert stats.memory_limit == 3899138048
    assert stats.memory_percent == 6.26


@pytest.mark.parametrize(
    "options", [None, SupervisorUpdateOptions(version="2024.01.0")]
)
async def test_supervisor_update(
    responses: aioresponses,
    supervisor_client: SupervisorClient,
    options: SupervisorUpdateOptions | None,
) -> None:
    """Test supervisor update API."""
    responses.post(f"{SUPERVISOR_URL}/supervisor/update", status=200)
    assert await supervisor_client.supervisor.update(options) is None
    assert responses.requests.keys() == {
        ("POST", URL(f"{SUPERVISOR_URL}/supervisor/update"))
    }


async def test_supervisor_reload(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test supervisor reload API."""
    responses.post(f"{SUPERVISOR_URL}/supervisor/reload", status=200)
    assert await supervisor_client.supervisor.reload() is None
    assert responses.requests.keys() == {
        ("POST", URL(f"{SUPERVISOR_URL}/supervisor/reload"))
    }


async def test_supervisor_restart(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test supervisor restart API."""
    responses.post(f"{SUPERVISOR_URL}/supervisor/restart", status=200)
    assert await supervisor_client.supervisor.restart() is None
    assert responses.requests.keys() == {
        ("POST", URL(f"{SUPERVISOR_URL}/supervisor/restart"))
    }


async def test_supervisor_options(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test supervisor options API."""
    responses.post(f"{SUPERVISOR_URL}/supervisor/options", status=200)
    assert (
        await supervisor_client.supervisor.set_options(
            SupervisorOptions(debug=True, debug_block=True)
        )
        is None
    )
    assert responses.requests.keys() == {
        ("POST", URL(f"{SUPERVISOR_URL}/supervisor/options"))
    }


async def test_supervisor_repair(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test supervisor repair API."""
    responses.post(f"{SUPERVISOR_URL}/supervisor/repair", status=200)
    assert await supervisor_client.supervisor.repair() is None
    assert responses.requests.keys() == {
        ("POST", URL(f"{SUPERVISOR_URL}/supervisor/repair"))
    }



"""Init file for aiohasupervisor."""

from aiohasupervisor.exceptions import (
    SupervisorAuthenticationError,
    SupervisorBadRequestError,
    SupervisorConnectionError,
    SupervisorError,
    SupervisorForbiddenError,
    SupervisorNotFoundError,
    SupervisorResponseError,
    SupervisorServiceUnavailableError,
    SupervisorTimeoutError,
)
from aiohasupervisor.root import SupervisorClient

__all__ = [
    "SupervisorError",
    "SupervisorConnectionError",
    "SupervisorAuthenticationError",
    "SupervisorBadRequestError",
    "SupervisorForbiddenError",
    "SupervisorNotFoundError",
    "SupervisorResponseError",
    "SupervisorServiceUnavailableError",
    "SupervisorTimeoutError",
    "SupervisorClient",
]



"""Models for OS APIs."""

from dataclasses import dataclass
from enum import StrEnum
from pathlib import PurePath

from .base import Options, Request, ResponseData

# --- ENUMS ----


class RaucState(StrEnum):
    """RaucState type."""

    GOOD = "good"
    BAD = "bad"
    ACTIVE = "active"


class BootSlotName(StrEnum):
    """BootSlotName type."""

    A = "A"
    B = "B"


# --- OBJECTS ----


@dataclass(frozen=True, slots=True)
class BootSlot(ResponseData):
    """BootSlot model."""

    state: str
    status: RaucState | None
    version: str | None


@dataclass(frozen=True, slots=True)
class OSInfo(ResponseData):
    """OSInfo model."""

    version: str | None
    version_latest: str | None
    update_available: bool
    board: str | None
    boot: str | None
    data_disk: str | None
    boot_slots: dict[str, BootSlot]


@dataclass(frozen=True, slots=True)
class OSUpdate(Request):
    """OSUpdate model."""

    version: str | None = None


@dataclass(frozen=True, slots=True)
class MigrateDataOptions(Request):
    """MigrateDataOptions model."""

    device: str


@dataclass(frozen=True, slots=True)
class DataDisk(ResponseData):
    """DataDisk model."""

    name: str
    vendor: str
    model: str
    serial: str
    size: int
    id: str
    dev_path: PurePath


@dataclass(frozen=True, slots=True)
class DataDiskList(ResponseData):
    """DataDiskList model."""

    disks: list[DataDisk]


@dataclass(frozen=True, slots=True)
class SetBootSlotOptions(Request):
    """SetBootSlotOptions model."""

    boot_slot: BootSlotName


@dataclass(frozen=True, slots=True)
class GreenInfo(ResponseData):
    """GreenInfo model."""

    activity_led: bool
    power_led: bool
    system_health_led: bool


@dataclass(frozen=True, slots=True)
class GreenOptions(Options):
    """GreenOptions model."""

    activity_led: bool | None = None
    power_led: bool | None = None
    system_health_led: bool | None = None


@dataclass(frozen=True, slots=True)
class YellowInfo(ResponseData):
    """YellowInfo model."""

    disk_led: bool
    heartbeat_led: bool
    power_led: bool


@dataclass(frozen=True, slots=True)
class YellowOptions(Options):
    """YellowOptions model."""

    disk_led: bool | None = None
    heartbeat_led: bool | None = None
    power_led: bool | None = None



"""Test resolution center supervisor client."""

from uuid import uuid4

from aioresponses import aioresponses
from yarl import URL

from aiohasupervisor import SupervisorClient
from aiohasupervisor.models import CheckOptions

from . import load_fixture
from .const import SUPERVISOR_URL


async def test_resolution_info(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test resolution center info API."""
    responses.get(
        f"{SUPERVISOR_URL}/resolution/info",
        status=200,
        body=load_fixture("resolution_info.json"),
    )
    info = await supervisor_client.resolution.info()
    assert info.checks[4].enabled is True
    assert info.checks[4].slug == "backups"
    assert info.issues[0].context == "system"
    assert info.issues[0].type == "no_current_backup"
    assert info.issues[0].reference is None
    assert info.issues[0].uuid.hex == "7f0eac2b61c9456dab6970507a276c36"
    assert info.suggestions[0].auto is False
    assert info.suggestions[0].context == "system"
    assert info.suggestions[0].type == "create_full_backup"
    assert info.suggestions[0].reference is None
    assert info.suggestions[0].uuid.hex == "f87d3556f02c4004a47111c072c76fac"
    assert info.unhealthy == ["supervisor"]
    assert info.unsupported == []


async def test_resolution_check_options(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test resolution center check options API."""
    responses.post(f"{SUPERVISOR_URL}/resolution/check/backups/options", status=200)
    assert (
        await supervisor_client.resolution.check_options(
            "backups", CheckOptions(enabled=False)
        )
        is None
    )
    assert responses.requests.keys() == {
        ("POST", URL(f"{SUPERVISOR_URL}/resolution/check/backups/options"))
    }


async def test_resolution_run_check(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test resolution center run check API."""
    responses.post(f"{SUPERVISOR_URL}/resolution/check/backups/run", status=200)
    assert await supervisor_client.resolution.run_check("backups") is None
    assert responses.requests.keys() == {
        ("POST", URL(f"{SUPERVISOR_URL}/resolution/check/backups/run"))
    }


async def test_resolution_apply_suggestion(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test resolution center apply suggestion API."""
    uuid = uuid4()
    responses.post(f"{SUPERVISOR_URL}/resolution/suggestion/{uuid.hex}", status=200)
    assert await supervisor_client.resolution.apply_suggestion(uuid) is None
    assert responses.requests.keys() == {
        ("POST", URL(f"{SUPERVISOR_URL}/resolution/suggestion/{uuid.hex}"))
    }


async def test_resolution_dismiss_suggestion(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test resolution center dismiss suggestion API."""
    uuid = uuid4()
    responses.delete(f"{SUPERVISOR_URL}/resolution/suggestion/{uuid.hex}", status=200)
    assert await supervisor_client.resolution.dismiss_suggestion(uuid) is None
    assert responses.requests.keys() == {
        ("DELETE", URL(f"{SUPERVISOR_URL}/resolution/suggestion/{uuid.hex}"))
    }


async def test_resolution_dismiss_issue(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test resolution center dismiss issue API."""
    uuid = uuid4()
    responses.delete(f"{SUPERVISOR_URL}/resolution/issue/{uuid.hex}", status=200)
    assert await supervisor_client.resolution.dismiss_issue(uuid) is None
    assert responses.requests.keys() == {
        ("DELETE", URL(f"{SUPERVISOR_URL}/resolution/issue/{uuid.hex}"))
    }


async def test_resolution_suggestions_for_issue(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test resolution center suggestions for issue API."""
    uuid = uuid4()
    responses.get(
        f"{SUPERVISOR_URL}/resolution/issue/{uuid.hex}/suggestions",
        status=200,
        body=load_fixture("resolution_suggestions_for_issue.json"),
    )
    result = await supervisor_client.resolution.suggestions_for_issue(uuid)
    assert result[0].auto is False
    assert result[0].context == "system"
    assert result[0].type == "create_full_backup"


async def test_resolution_healthcheck(
    responses: aioresponses, supervisor_client: SupervisorClient
) -> None:
    """Test resolution center healthcheck API."""
    responses.post(f"{SUPERVISOR_URL}/resolution/healthcheck", status=200)
    assert await supervisor_client.resolution.healthcheck() is None
    assert responses.requests.keys() == {
        ("POST", URL(f"{SUPERVISOR_URL}/resolution/healthcheck"))
    }
